#include "StdAfx.h"
#include "AG_JavaBytecodeGenerator.h"

namespace TinyPascal { namespace Emitter {
const CAG_JavaBytecodeGenerator::u4 CAG_JavaBytecodeGenerator::magic = 0xCAFEBABE;

CAG_JavaBytecodeGenerator::CAG_JavaBytecodeGenerator(void)
{
	// Determine endianess of the machine
	unsigned int x   = 0x00010203;
	unsigned char *p = (unsigned char*)&x;
	b0 = p[0];b1 = p[1];b2 = p[2];b3 = p[3];
}

CAG_JavaBytecodeGenerator::~CAG_JavaBytecodeGenerator(void)
{
	// Clean up constant_pool
	for(size_t i=0; i<constant_pool.size(); ++i)
		delete constant_pool[i];

	// Clean up fields
	for(size_t i=0; i<fields.size(); ++i)
		delete fields[i];

	// Clean up methods
	for(size_t i=0; i<methods.size(); ++i)
		delete methods[i];

	// Clean up attributes vector
	for(size_t i=0; i<attributes.size(); ++i)
		delete attributes[i];

	// Clean up class table reference
	m_Class_Table.clear();

	// Clean up Utf8 table
	m_Utf8_Table.clear();
}

void CAG_JavaBytecodeGenerator::Initialize()
{
	// Clean up constant_pool
	for(size_t i=0; i<constant_pool.size(); ++i)
		delete constant_pool[i];

	// Clean up fields
	for(size_t i=0; i<fields.size(); ++i)
		delete fields[i];

	// Clean up methods
	for(size_t i=0; i<methods.size(); ++i)
		delete methods[i];

	// Clean up attributes vector
	for(size_t i=0; i<attributes.size(); ++i)
		delete attributes[i];

	// Clean up class table reference
	m_Class_Table.clear();

	// Clean up Utf8 table
	m_Utf8_Table.clear();

	// Clean up Field reference table
	m_Field_Table.clear();

	// Clean up integer factor table
	m_Integer_Table.clear();

	// Cleant up float table
	m_Float_Table.clear();

	// Create constant_pool entry 0 as specified by JVM specs
	constant_pool.push_back(new cp_info());

	// Set version to 1.5
	major_version.v = 49;
	minor_version.v = 0;

	// Set access flags to be public
	access_flags.v = 0x0001;

	// Set this class index
	this_class = AddClassInfo(m_strClassName);

	// Set super class index
	super_class = AddClassInfo("java/lang/Object");

	// Add reference to System.out.println() method
	m_PrintlnRef = AddMethodrefInfo("java/io/PrintStream", "println", "()V");

	// Add reference to System.out.println(int) method
	m_PrintlnIntRef = AddMethodrefInfo("java/io/PrintStream", "println", "(I)V");

	// Add reference to System.out.println(float) method
	m_PrintlnFloatRef = AddMethodrefInfo("java/io/PrintStream", "println", "(F)V");

	// We must add reference to field where static println methods can be found
	m_PrintStreamRef = AddFieldrefInfo("java/lang/System", "out", "Ljava/io/PrintStream;");

	// Add defaull constructor
	AddDefaultConstructor();

	// Add main method signature
	AddMainMethod();
}

void CAG_JavaBytecodeGenerator::AddDefaultConstructor()
{
	// Add Method
	method_info* pMethodInfo				= new method_info();
	pMethodInfo->access_flags				= 0x0001;
	pMethodInfo->name_index					= AddUtf8Info("<init>");
	pMethodInfo->descriptor_index			= AddUtf8Info("()V");

	// Add Code Attributed for Method
	Code_attribute* pCodeAttribute			= new Code_attribute();
	pCodeAttribute->attribute_name_index	= AddUtf8Info("Code");
	pCodeAttribute->max_stack				= 1;
	pCodeAttribute->max_locals				= 1;

	pCodeAttribute->code.push_back(aload_0);
	pCodeAttribute->code.push_back(invokespecial);
	
	// Construct invokespecial parameter - Methodref Info structure
	u2 index = AddMethodrefInfo("java/lang/Object", "<init>", "()V");
	Emit(index, pCodeAttribute->code);

	pCodeAttribute->code.push_back(_return);

	pMethodInfo->attributes.push_back(pCodeAttribute);
	methods.push_back(pMethodInfo);
}

void CAG_JavaBytecodeGenerator::AddMainMethod()
{
	// Add main method of the class -> here is where everything happens
	method_info* pMethodInfo		= new method_info();
	pMethodInfo->access_flags		= 0x0001 | 0x0008; 
	pMethodInfo->name_index			= AddUtf8Info("main");
	pMethodInfo->descriptor_index	= AddUtf8Info("([Ljava/lang/String;)V");
	m_pMainMethodCode							= new Code_attribute();
	m_pMainMethodCode->attribute_name_index		= AddUtf8Info("Code");
	
	pMethodInfo->attributes.push_back(m_pMainMethodCode);
	methods.push_back(pMethodInfo);
}
CAG_JavaBytecodeGenerator::u2 CAG_JavaBytecodeGenerator::AddUtf8Info(string bytes)
{
	u2 index;

	// first check if we have such string in hash_map already
	if(m_Utf8_Table.find(bytes) != m_Utf8_Table.end())
		return m_Utf8_Table[bytes];

	// Add new utf 8 info string
	CONSTANT_Utf8_Info* pUtf8Info = new CONSTANT_Utf8_Info(bytes);
	index = constant_pool.size();
	constant_pool.push_back(pUtf8Info);
	m_Utf8_Table[bytes] = index;

	return index;
}

CAG_JavaBytecodeGenerator::u2 CAG_JavaBytecodeGenerator::AddClassInfo(string name)
{
	u2 index;

	// Check does the class already exists
	if(m_Class_Table.find(name) != m_Class_Table.end())
		return m_Class_Table[name];

	// Class does not exist so add new one
	CONSTANT_Class_Info* pClassInfo = new CONSTANT_Class_Info(AddUtf8Info(name));
	index = constant_pool.size();
	constant_pool.push_back(pClassInfo);
	m_Class_Table[name] = index;

	return index;
}

CAG_JavaBytecodeGenerator::u2 CAG_JavaBytecodeGenerator::AddNameAndTypeInfo(string name, string descriptor)
{
	u2 index;

	CONSTANT_NameAndType_Info* pNameAndTypeInfo = new CONSTANT_NameAndType_Info();
	pNameAndTypeInfo->name_index = AddUtf8Info(name);
	pNameAndTypeInfo->descriptor_index = AddUtf8Info(descriptor);
	index = constant_pool.size();
	constant_pool.push_back(pNameAndTypeInfo);

	return index;
}

CAG_JavaBytecodeGenerator::u2 CAG_JavaBytecodeGenerator::AddMethodrefInfo(string _class, string name, string type)
{
	u2 index;

	CONSTANT_Methodref_Info* pMethodRef		= new CONSTANT_Methodref_Info();
	pMethodRef->class_index					= AddClassInfo(_class);
	pMethodRef->name_and_type_index			= AddNameAndTypeInfo(name, type);
	index = constant_pool.size();
	constant_pool.push_back(pMethodRef);

	return index;
}

CAG_JavaBytecodeGenerator::u2 CAG_JavaBytecodeGenerator::AddFieldrefInfo(string _class, string name, string type)
{
	u2 index;

	if(m_Field_Table.find(_class+name+type) != m_Field_Table.end())
		return m_Field_Table[_class+name+type];

	CONSTANT_Fieldref_Info* pFieldRef		= new CONSTANT_Fieldref_Info();
	pFieldRef->class_index					= AddClassInfo(_class);
	pFieldRef->name_and_type_index			= AddNameAndTypeInfo(name, type);
	index = constant_pool.size();
	constant_pool.push_back(pFieldRef);
	m_Field_Table[_class+name+type] = index;

	return index;
}

CAG_JavaBytecodeGenerator::u2 CAG_JavaBytecodeGenerator::AddIntegerInfo(signed int i)
{
	u2 index;

	// add this integer factor to constant_pool, unless already present
	if(m_Integer_Table.find(i) != m_Integer_Table.end())
		return m_Integer_Table[i];

	CONSTANT_Integer_Info* pIntInfo = new CONSTANT_Integer_Info();
	pIntInfo->bytes.i = i;
	index = constant_pool.size();
	constant_pool.push_back(pIntInfo);
	m_Integer_Table[i] = index;

	return index;
}

CAG_JavaBytecodeGenerator::u2 CAG_JavaBytecodeGenerator::AddFloatInfo(float f)
{
	u2 index;

	// add this integer factor to constant_pool, unless already present
	if(m_Float_Table.find(f) != m_Float_Table.end())
		return m_Float_Table[f];

	CONSTANT_Float_Info* pFloatInfo = new CONSTANT_Float_Info();
	pFloatInfo->bytes.f = f;
	index = constant_pool.size();
	constant_pool.push_back(pFloatInfo);
	m_Float_Table[f] = index;

	return index;
}

void CAG_JavaBytecodeGenerator::AddField(string name, string type)
{
	field_info* pFieldInfo = new field_info();
	// Access flags are always (in our case) public static
	pFieldInfo->access_flags = 0x0001 | 0x0008; 
	pFieldInfo->name_index = AddUtf8Info(name);
	pFieldInfo->descriptor_index = AddUtf8Info(type);
	fields.push_back(pFieldInfo);
}
void CAG_JavaBytecodeGenerator::GetBytecode(cp_info* p, string& program)
{
	switch(p->tag.b[0])
	{
	case cp_info::CONSTANT_Class:
		GetBytecode((CONSTANT_Class_Info*)p, program);
		break;
	case cp_info::CONSTANT_Utf8:
		GetBytecode((CONSTANT_Utf8_Info*)p, program);
		break;
	case cp_info::CONSTANT_Methodref:
		GetBytecode((CONSTANT_Methodref_Info*)p, program);
		break;
	case cp_info::CONSTANT_Fieldref:
		GetBytecode((CONSTANT_Fieldref_Info*)p, program);
		break;
	case cp_info::CONSTANT_NameAndType:
		GetBytecode((CONSTANT_NameAndType_Info*)p, program);
		break;
	case cp_info::CONSTANT_Integer:
		GetBytecode((CONSTANT_Integer_Info*)p, program);
		break;
	case cp_info::CONSTANT_Float:
		GetBytecode((CONSTANT_Float_Info*)p, program);
		break;
	default: Emit(p->tag, program);
	};
}

void CAG_JavaBytecodeGenerator::GetBytecode(CONSTANT_Class_Info* p, string& program)
{
	Emit(p->tag, program);
	Emit(p->name_index, program);
}

void CAG_JavaBytecodeGenerator::GetBytecode(CONSTANT_Methodref_Info* p, string& program)
{
	Emit(p->tag, program);
	Emit(p->class_index, program);
	Emit(p->name_and_type_index, program);
}

void CAG_JavaBytecodeGenerator::GetBytecode(CONSTANT_Fieldref_Info* p, string& program)
{
	Emit(p->tag, program);
	Emit(p->class_index, program);
	Emit(p->name_and_type_index, program);
}

void CAG_JavaBytecodeGenerator::GetBytecode(CONSTANT_NameAndType_Info* p, string& program)
{
	Emit(p->tag, program);
	Emit(p->name_index, program);
	Emit(p->descriptor_index, program);
}

void CAG_JavaBytecodeGenerator::GetBytecode(CONSTANT_Utf8_Info* p, string& program)
{
	Emit(p->tag, program);
	p->length = p->bytes.size();
	Emit(p->length, program);
	program += p->bytes;
}

void CAG_JavaBytecodeGenerator::GetBytecode(CONSTANT_Integer_Info* p, string& program)
{
	Emit(p->tag, program);
	Emit(p->bytes, program);
}

void CAG_JavaBytecodeGenerator::GetBytecode(CONSTANT_Float_Info* p, string& program)
{
	Emit(p->tag, program);
	Emit(p->bytes, program);
}

void CAG_JavaBytecodeGenerator::GetBytecode(method_info* p, string& program)
{
	Emit(p->access_flags, program);
	Emit(p->name_index, program);
	Emit(p->descriptor_index, program);
	p->attributes_count = p->attributes.size();
	Emit(p->attributes_count, program);
	for(size_t i=0; i<p->attributes.size(); ++i)
		program += p->attributes[i]->GetBytecode();
}

void CAG_JavaBytecodeGenerator::GetBytecode(field_info* p, string& program)
{
	Emit(p->access_flags, program);
	Emit(p->name_index, program);
	Emit(p->descriptor_index, program);
	p->attributes_count = p->attributes.size();
	Emit(p->attributes_count, program);
	for(size_t i=0; i<p->attributes.size(); ++i)
		program += p->attributes[i]->GetBytecode();
}

string CAG_JavaBytecodeGenerator::GetProgram()
{
	string strRes;

	Emit(magic, strRes);
	Emit(minor_version, strRes);
	Emit(major_version, strRes);
	constant_pool_count.v = constant_pool.size();
	Emit(constant_pool_count, strRes);
	for(size_t i=1; i<constant_pool.size(); ++i)
		GetBytecode(constant_pool[i], strRes);
	
	Emit(access_flags, strRes);
	Emit(this_class, strRes);
	Emit(super_class, strRes);

	interfaces_count.v = interfaces.size();
	Emit(interfaces_count, strRes);
	for(size_t i=0; i<interfaces.size(); ++i)
		Emit(interfaces[i], strRes);

	fields_count.v = fields.size();
	Emit(fields_count, strRes);
	for(size_t i=0; i<fields.size(); ++i)
		GetBytecode(fields[i], strRes);

	methods_count.v = methods.size();
	Emit(methods_count, strRes);
	for(size_t i=0; i<methods.size(); ++i)
		GetBytecode(methods[i], strRes);
	
	attributes_count.v = attributes.size();
	Emit(attributes_count, strRes);
	for(size_t i=0; i<attributes.size(); ++i)
		strRes += attributes[i]->GetBytecode();

	return strRes;
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool CAG_JavaBytecodeGenerator::EmitVarDecl(CAG_TPSymbolTable::SymbolTableType& SymbolTable)
{
	/*
		Emitting varible declaration is nothing else but doing following:
		1. Create field of the corresponding type
		2. Create reference to that field for later use

		-> Ideas for optimization:
		1. Check if the field is used and then create field/reference
	*/
	CAG_TPSymbolTable::SymbolTableType::const_iterator iter = SymbolTable.begin();
	for(iter; iter != SymbolTable.end(); ++iter)
	{
		string strType;
		if(iter->second == CAG_TPSymbolTable::real)
			strType = "F";
		else if(iter->second == CAG_TPSymbolTable::integer)
			strType = "I";

		AddField("_" + iter->first, strType);
		AddFieldrefInfo(m_strClassName, "_" + iter->first, strType);
	}

	return true;
}


bool CAG_JavaBytecodeGenerator::PushIdentifier(CAG_TPSymbolTable::TreeNodeType& Identifier)
{
	// Get identifier
	string strIdentifier(Identifier.value.begin(),
						 Identifier.value.end());
	strIdentifier = "_" + strIdentifier;
	u2 index;

	// Look up the index in constant pool table
	if(Identifier.value.value() == CAG_TPSymbolTable::integer)
		index = m_Field_Table[m_strClassName+strIdentifier+"I"];
	else index = m_Field_Table[m_strClassName+strIdentifier+"F"];

	// Push field value on the stack
	ComputeStack(true, m_pMainMethodCode->max_stack);
	m_pMainMethodCode->code.push_back(getstatic);
	Emit(index, m_pMainMethodCode->code);

	return true;
}

bool CAG_JavaBytecodeGenerator::PushIntegerFactor(CAG_TPSymbolTable::TreeNodeType& IntegerFactor)
{
	// Get Factor
	string strIntegerFactor(IntegerFactor.value.begin(),
						    IntegerFactor.value.end());
	
	ComputeStack(true, m_pMainMethodCode->max_stack);
	m_pMainMethodCode->code.push_back(ldc_w);
	Emit(AddIntegerInfo(atoi(strIntegerFactor.c_str())), m_pMainMethodCode->code);

	return true;
}

bool CAG_JavaBytecodeGenerator::PushRealFactor(CAG_TPSymbolTable::TreeNodeType& RealFactor)
{
	// Get Factor
	string strRealFactor(RealFactor.value.begin(),
				 	     RealFactor.value.end());

	ComputeStack(true, m_pMainMethodCode->max_stack);
	m_pMainMethodCode->code.push_back(ldc_w);
	Emit(AddFloatInfo(atof(strRealFactor.c_str())), m_pMainMethodCode->code);

	return true;
}

bool CAG_JavaBytecodeGenerator::EmitAssignment(CAG_TPSymbolTable::TreeChildrenType& OpAssign)
{
	/*
		Assignment is emitted in two steps:
		1. First emit right-hand side of the assignment
		2. Then pop the result from stack and store it into left-hand side
	*/
	switch(OpAssign[1].value.id().to_long())
	{
	case CAG_TinyPascalGrammar::ExpressionID:
		EmitExpression(OpAssign[1]);
		break;
	case CAG_TinyPascalGrammar::TermID:
		EmitTerm(OpAssign[1]);
		break;
	case CAG_TinyPascalGrammar::RealFactorID:
		PushRealFactor(OpAssign[1]);
		break;
	case CAG_TinyPascalGrammar::IntegerFactorID:
		PushIntegerFactor(OpAssign[1]);
		break;
	case CAG_TinyPascalGrammar::IdentifierID:
		PushIdentifier(OpAssign[1]);
		break;
	}

	// Get left hand side variable
	string strIdentifier(OpAssign[0].value.begin(),
						 OpAssign[0].value.end());
	if(OpAssign[0].value.value() == CAG_TPSymbolTable::integer)
		strIdentifier = m_strClassName + "_" + strIdentifier + "I";
	else strIdentifier = m_strClassName + "_" + strIdentifier + "F";
	u2 index = m_Field_Table[strIdentifier];

	// Pop the result from stack and store it into identifier
	if(OpAssign[1].value.value() == CAG_TPSymbolTable::integer)
	{
		if(OpAssign[0].value.value() == CAG_TPSymbolTable::integer)
		{
			// Integer := Integer
			ComputeStack(false, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(putstatic);
			Emit(index, m_pMainMethodCode->code);
		}
		else
		{
			// Real := Integer -> we have to convert integer to real
			ComputeStack(true, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(i2f);
			ComputeStack(false, m_pMainMethodCode->max_stack);
			ComputeStack(false, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(putstatic);
			Emit(index, m_pMainMethodCode->code);
		}
	}
	else
	{
		if(OpAssign[0].value.value() == CAG_TPSymbolTable::integer)
		{
			// Integer := Real -> we have to convert real to integer
			ComputeStack(true, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(f2i);
			ComputeStack(false, m_pMainMethodCode->max_stack);
			ComputeStack(false, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(putstatic);
			Emit(index, m_pMainMethodCode->code);
		}
		else
		{
			// Real := Real
			ComputeStack(false, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(putstatic);
			Emit(index, m_pMainMethodCode->code);
		}
	}

	return true;
}

bool CAG_JavaBytecodeGenerator::EmitTerm(CAG_TPSymbolTable::TreeNodeType& Term)
{
	/*
		Every term has at most two children. Once we evaluate both children,
		we are ready to evaluate the term.
	*/
	for(size_t i=0; i<Term.children.size(); ++i)
	{
		switch(Term.children[i].value.id().to_long())
		{
			case CAG_TinyPascalGrammar::TermID:
				EmitTerm(Term.children[i]);
				break;
			case CAG_TinyPascalGrammar::ExpressionID:
				EmitExpression(Term.children[i]);
				break;
			case CAG_TinyPascalGrammar::RealFactorID:
				PushRealFactor(Term.children[i]);
				break;
			case CAG_TinyPascalGrammar::IntegerFactorID:
				PushIntegerFactor(Term.children[i]);
				break;
			case CAG_TinyPascalGrammar::IdentifierID:
				PushIdentifier(Term.children[i]);
				break;
		}

		/* 
			Check if we need to do type conversion!
			THIS CASE SHOULD NOT HAPPEN THOUGH! THIS IS BECAUSE IT WOULD BE
			WRONG TO EVALUATE TYPE OF TERM TO BE INTEGER IF ANY OPERANDS ARE
			FLOAT -> THIS MEANS THAT OUR TYPE DERIVATION DID NOT WORK PROPERLY!
		*/
		if(Term.value.value() == CAG_TPSymbolTable::integer && 
		   Term.children[i].value.value() == CAG_TPSymbolTable::real)
		{
			ComputeStack(true, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(f2i);
			ComputeStack(false, m_pMainMethodCode->max_stack);
		}

		/* This is only thing we need to check really -> factor integer, but term real */
		if(Term.value.value() == CAG_TPSymbolTable::real && 
		   Term.children[i].value.value() == CAG_TPSymbolTable::integer)
		{
			ComputeStack(true, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(i2f);
			ComputeStack(false, m_pMainMethodCode->max_stack);
		}
	}

	/* Check what it is */
	string strOp(Term.value.begin(),
		         Term.value.end());
	if(strOp == "*")
		EmitMultiplication(Term);
	else EmitDivision(Term);

	return true;
}

bool CAG_JavaBytecodeGenerator::EmitExpression(CAG_TPSymbolTable::TreeNodeType& Expression)
{
	/*
		Every term has at most two children. Once we evaluate both children,
		we are ready to evaluate the term.
	*/
	for(size_t i=0; i<Expression.children.size(); ++i)
	{
		switch(Expression.children[i].value.id().to_long())
		{
			case CAG_TinyPascalGrammar::TermID:
				EmitTerm(Expression.children[i]);
				break;
			case CAG_TinyPascalGrammar::ExpressionID:
				EmitExpression(Expression.children[i]);
				break;
			case CAG_TinyPascalGrammar::RealFactorID:
				PushRealFactor(Expression.children[i]);
				break;
			case CAG_TinyPascalGrammar::IntegerFactorID:
				PushIntegerFactor(Expression.children[i]);
				break;
			case CAG_TinyPascalGrammar::IdentifierID:
				PushIdentifier(Expression.children[i]);
				break;
		}

		/* 
			Check if we need to do type conversion!
			THIS CASE SHOULD NOT HAPPEN THOUGH! THIS IS BECAUSE IT WOULD BE
			WRONG TO EVALUATE TYPE OF TERM TO BE INTEGER IF ANY OPERANDS ARE
			FLOAT -> THIS MEANS THAT OUR TYPE DERIVATION DID NOT WORK PROPERLY!
		*/
		if(Expression.value.value() == CAG_TPSymbolTable::integer && 
		   Expression.children[i].value.value() == CAG_TPSymbolTable::real)
		{
			ComputeStack(true, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(f2i);
			ComputeStack(false, m_pMainMethodCode->max_stack);
		}

		/* This is only thing we need to check really -> factor integer, but term real */
		if(Expression.value.value() == CAG_TPSymbolTable::real && 
		   Expression.children[i].value.value() == CAG_TPSymbolTable::integer)
		{
			ComputeStack(true, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(i2f);
			ComputeStack(false, m_pMainMethodCode->max_stack);
		}
	}

	/* Check what it is */
	string strOp(Expression.value.begin(),
		         Expression.value.end());
	if(strOp == "+")
		EmitAddition(Expression);
	else EmitSubtraction(Expression);

	return true;
}

bool CAG_JavaBytecodeGenerator::EmitAddition(CAG_TPSymbolTable::TreeNodeType& Addition)
{
	/* 
		Addition is done by popping both operands from stack 
		and adding them. Result is pushed onto stack. This operation
		must be performed with type conversion in mind.
	*/
	if(Addition.value.value() == CAG_TPSymbolTable::integer)
	{
		/*
			Here we perform integer addition
		*/
		ComputeStack(false, m_pMainMethodCode->max_stack);
		ComputeStack(false, m_pMainMethodCode->max_stack);
		m_pMainMethodCode->code.push_back(iadd);
		ComputeStack(true, m_pMainMethodCode->max_stack);
	}
	else
	{
		/*
			Here we perform float addition
		*/
		ComputeStack(false, m_pMainMethodCode->max_stack);
		ComputeStack(false, m_pMainMethodCode->max_stack);
		m_pMainMethodCode->code.push_back(fadd);
		ComputeStack(true, m_pMainMethodCode->max_stack);
	}

	return true;
}

bool CAG_JavaBytecodeGenerator::EmitSubtraction(CAG_TPSymbolTable::TreeNodeType& Subtraction)
{
	/* 
		Subtraction is done by popping both operands from stack 
		and subtracting them. Result is pushed onto stack. This operation
		must be performed with type conversion in mind.
	*/
	if(Subtraction.value.value() == CAG_TPSymbolTable::integer)
	{
		/*
			Here we perform integer addition
		*/
		ComputeStack(false, m_pMainMethodCode->max_stack);
		ComputeStack(false, m_pMainMethodCode->max_stack);
		m_pMainMethodCode->code.push_back(isub);
		ComputeStack(true, m_pMainMethodCode->max_stack);
	}
	else
	{
		/*
			Here we perform float addition
		*/
		ComputeStack(false, m_pMainMethodCode->max_stack);
		ComputeStack(false, m_pMainMethodCode->max_stack);
		m_pMainMethodCode->code.push_back(fsub);
		ComputeStack(true, m_pMainMethodCode->max_stack);	
	}

	return true;
}

bool CAG_JavaBytecodeGenerator::EmitDivision(CAG_TPSymbolTable::TreeNodeType& Division)
{
	/* 
		Division is done by popping both operands from stack 
		and dividing them. Result is pushed onto stack. This operation
		must be performed with type conversion in mind.
	*/
	if(Division.value.value() == CAG_TPSymbolTable::integer)
	{
		/*
			Here we perform integer division
		*/
		ComputeStack(false, m_pMainMethodCode->max_stack);
		ComputeStack(false, m_pMainMethodCode->max_stack);
		m_pMainMethodCode->code.push_back(idiv);
		ComputeStack(true, m_pMainMethodCode->max_stack);
	}
	else
	{
		/*
			Here we perform float multiplication
		*/
		ComputeStack(false, m_pMainMethodCode->max_stack);
		ComputeStack(false, m_pMainMethodCode->max_stack);
		m_pMainMethodCode->code.push_back(fdiv);
		ComputeStack(true, m_pMainMethodCode->max_stack);
	}

	return true;
}

bool CAG_JavaBytecodeGenerator::EmitMultiplication(CAG_TPSymbolTable::TreeNodeType& Multiplication)
{
	/* Multiplication is done by popping both operands from stack 
	   and multiplying them. Result is pushed onto stack. This operation
	   must be performed with type conversion in mind
	*/
	if(Multiplication.value.value() == CAG_TPSymbolTable::integer)
	{
		/*
			Here we perform integer multiplication
		*/
		ComputeStack(false, m_pMainMethodCode->max_stack);
		ComputeStack(false, m_pMainMethodCode->max_stack);
		m_pMainMethodCode->code.push_back(imul);
		ComputeStack(true, m_pMainMethodCode->max_stack);
	}
	else
	{
		/*
			Here we perform float multiplication
		*/
		ComputeStack(false, m_pMainMethodCode->max_stack);
		ComputeStack(false, m_pMainMethodCode->max_stack);
		m_pMainMethodCode->code.push_back(fmul);
		ComputeStack(true, m_pMainMethodCode->max_stack);
	}

	return true;
}

bool CAG_JavaBytecodeGenerator::EmitWritelnStatement(CAG_TPSymbolTable::TreeChildrenType& WritelnStmt)
{
	// Handle special case where writeln is empty
	if(WritelnStmt.size() == 0)
	{
		ComputeStack(true, m_pMainMethodCode->max_stack);
		m_pMainMethodCode->code.push_back(getstatic);
		Emit(m_PrintStreamRef, m_pMainMethodCode->code);
		ComputeStack(false, m_pMainMethodCode->max_stack);
		m_pMainMethodCode->code.push_back(invokevirtual);
		Emit(m_PrintlnRef, m_pMainMethodCode->code);
	}

	for(size_t i=0; i<WritelnStmt.size(); ++i)
	{
		ComputeStack(true, m_pMainMethodCode->max_stack);
		m_pMainMethodCode->code.push_back(getstatic);
		Emit(m_PrintStreamRef, m_pMainMethodCode->code);

		switch(WritelnStmt[i].value.id().to_long())
		{
		case CAG_TinyPascalGrammar::ExpressionID:
			EmitExpression(WritelnStmt[i]);
			break;
		case CAG_TinyPascalGrammar::TermID:
			EmitTerm(WritelnStmt[i]);
			break;
		case CAG_TinyPascalGrammar::RealFactorID:
			PushRealFactor(WritelnStmt[i]);
			break;
		case CAG_TinyPascalGrammar::IntegerFactorID:
			PushIntegerFactor(WritelnStmt[i]);
			break;
		case CAG_TinyPascalGrammar::IdentifierID:
			PushIdentifier(WritelnStmt[i]);
			break;
		}

		// Pop result from the stack and print either real or integer
		if(WritelnStmt[i].value.value() == CAG_TPSymbolTable::integer)
		{
			// This pops 2 operands from the stack
			ComputeStack(false, m_pMainMethodCode->max_stack);
			ComputeStack(false, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(invokevirtual);
			Emit(m_PrintlnIntRef, m_pMainMethodCode->code);
		}
		else
		{
			// This pops 2 operands from the stack
			ComputeStack(false, m_pMainMethodCode->max_stack);
			ComputeStack(false, m_pMainMethodCode->max_stack);
			m_pMainMethodCode->code.push_back(invokevirtual);
			Emit(m_PrintlnFloatRef, m_pMainMethodCode->code);
		}
	}

	return true;
}

bool CAG_JavaBytecodeGenerator::EmitMainStatementPart(CAG_TPSymbolTable::TreeChildrenType& MainStatementPart)
{
	/*
		MainStatementPart is always surrounded by begin and end so that first
		and last child could be ommitted during traversal. Following loop emits
		all children of the MainStatementPart
	*/
	for(size_t i=1; i<MainStatementPart.size()-1; ++i)
	{
		switch(MainStatementPart[i].value.id().to_long())
		{
		case CAG_TinyPascalGrammar::OpAssignID:
			EmitAssignment(MainStatementPart[i].children);
			break;
		case CAG_TinyPascalGrammar::MainStatementPartID:
			EmitMainStatementPart(MainStatementPart[i].children);
			break;
		case CAG_TinyPascalGrammar::WritelnStmtID:
			EmitWritelnStatement(MainStatementPart[i].children);
			break;
		case CAG_TinyPascalGrammar::ForStatementID:
			EmitForLoop(MainStatementPart[i].children);
			break;
		}
	}

	return true;
}

bool CAG_JavaBytecodeGenerator::EmitForLoop(CAG_TPSymbolTable::TreeChildrenType& ForLoop)
{
	/*
		For loop is emitted in following way:
		
		1.		Evaluate(OpAssign) - LoopVariable contains initial value
		2.		Store byte position for later reference 
		3.		Evaluate(Expression) - result will be pushed on stack
		4.		Compare(Expression, LoopVariable) - if true then jump to end position
					- If keyword "downto" is used then comparison should be LoopVariable >= Expression
					- If keyword "to" is used then comparison should be LoopVariable <= Expression.
		5.		Evaluate(Statement)
		6.		Modify LoopVariable
					- If keyword "downto" is used then Decrease(LoopVariable)
					- If keyword "to" is used then Increase(LoopVariable)
		7.		Jump to stored position determined in (2)
		8.		This is end position

		NOTE: Such loop interpretation is different than "standard" Pascal, where loop
		variable cannot be changed inside the loop.

	*/

	// Get index of the loop variable
	string strLoopVar(ForLoop[1].children[0].value.begin(),
					  ForLoop[1].children[0].value.end());
	strLoopVar = m_strClassName + "_" + strLoopVar + "I";
	u2 LoopVarIndex = m_Field_Table[strLoopVar];

	// Determine if we are increasing or decreasing
	bool bIncrease = true;
	string strKeyword(ForLoop[2].value.begin(),
					  ForLoop[2].value.end());
	transform(strKeyword.begin(), strKeyword.end(), strKeyword.begin(), tolower);
	if(strKeyword.compare("downto") == 0)
		bIncrease = false;

	// Evaluate assignment
	EmitAssignment(ForLoop[1].children);

	// Determine position for jump
	u2 LoopOffset = m_pMainMethodCode->code.size();

	// Evaluate Expression Part
	switch(ForLoop[3].value.id().to_long())
	{
	case CAG_TinyPascalGrammar::ExpressionID:
		EmitExpression(ForLoop[3]);
		break;
	case CAG_TinyPascalGrammar::TermID:
		EmitTerm(ForLoop[3]);
		break;
	case CAG_TinyPascalGrammar::RealFactorID:
		PushRealFactor(ForLoop[3]);
		break;
	case CAG_TinyPascalGrammar::IntegerFactorID:
		PushIntegerFactor(ForLoop[3]);
		break;
	case CAG_TinyPascalGrammar::IdentifierID:
		PushIdentifier(ForLoop[3]);
		break;
	}

	/*
		Comparison is perfomed in following way:
		1. Result from expression part is on the stack already
		2. Push loop variable value onto the stack 
		3. Compare using either if_icmplt or if_icmpgt 
		
		Here we have value1 = expression result and value2 = loop variable value
	*/
	ComputeStack(true, m_pMainMethodCode->max_stack);
	m_pMainMethodCode->code.push_back(getstatic);
	Emit(LoopVarIndex, m_pMainMethodCode->code);
	size_t LoopEndPointer;
	if(bIncrease)
		m_pMainMethodCode->code.push_back(if_icmplt);
	else m_pMainMethodCode->code.push_back(if_icmpgt);
	
	// Emit 0 for now, we'll change this later
	LoopEndPointer = m_pMainMethodCode->code.size();
	Emit(u2(0), m_pMainMethodCode->code);
	ComputeStack(false, m_pMainMethodCode->max_stack);
	ComputeStack(false, m_pMainMethodCode->max_stack);

	// Evaluate loop statement part
	if(ForLoop.size() == 6)
	{
		switch(ForLoop[5].value.id().to_long())
		{
		case CAG_TinyPascalGrammar::OpAssignID:
			EmitAssignment(ForLoop[5].children);
			break;
		case CAG_TinyPascalGrammar::MainStatementPartID:
			EmitMainStatementPart(ForLoop[5].children);
			break;
		case CAG_TinyPascalGrammar::WritelnStmtID:
			EmitWritelnStatement(ForLoop[5].children);
			break;
		case CAG_TinyPascalGrammar::ForStatementID:
			EmitForLoop(ForLoop[5].children);
			break;
		}
	}

	// Increase/Decrease loop variable using "iadd" instruction
	ComputeStack(true, m_pMainMethodCode->max_stack);
	m_pMainMethodCode->code.push_back(getstatic);
	Emit(LoopVarIndex, m_pMainMethodCode->code);
	ComputeStack(true, m_pMainMethodCode->max_stack);
	if(bIncrease)
		m_pMainMethodCode->code.push_back(iconst_1);
	else m_pMainMethodCode->code.push_back(iconst_m1);
	m_pMainMethodCode->code.push_back(iadd);
	ComputeStack(false, m_pMainMethodCode->max_stack);
	m_pMainMethodCode->code.push_back(putstatic);
	Emit(LoopVarIndex, m_pMainMethodCode->code);
	ComputeStack(false, m_pMainMethodCode->max_stack);

	// Evaluate loop again
	LoopOffset.s = LoopOffset.s - m_pMainMethodCode->code.size();
	m_pMainMethodCode->code.push_back(_goto);
	Emit(LoopOffset, m_pMainMethodCode->code);

	// Now we need to adjust the offset to be able to jump to the end of the loop upon successful comparison
	u2 EndOffset; EndOffset.s = m_pMainMethodCode->code.size() - LoopEndPointer + 1;
	m_pMainMethodCode->code[LoopEndPointer] = EndOffset.b[b0 & 0x01];
	m_pMainMethodCode->code[LoopEndPointer+1] = EndOffset.b[b1 & 0x01];

	return true;
}


bool CAG_JavaBytecodeGenerator::Emit(CAG_TPSymbolTable::TreeChildrenType& AST,
		                             CAG_TPSymbolTable::SymbolTableType& SymbolTable)
{
	// Write program name as comment
	m_strClassName = "_" + string(AST[0].children[0].value.begin(), AST[0].children[0].value.end());

	// Clean up all resources first
	Initialize();

	// Reinitialize main method stack computation
	ComputeStack(false, m_pMainMethodCode->max_stack, true);

	// Define all variables
	EmitVarDecl(SymbolTable);

	// Recursivly emit main statement part
	// Find main statement part subtree
	CAG_TPSymbolTable::TreeChildrenType& MainStatementPart = CAG_TPSymbolTable::TreeChildrenType();
	for(size_t i=0; i<AST[0].children.size(); ++i)
		if(AST[0].children[i].value.id().to_long() == CAG_TinyPascalGrammar::MainStatementPartID)
			MainStatementPart = AST[0].children[i].children;
	EmitMainStatementPart(MainStatementPart);

	m_pMainMethodCode->max_locals = 1;

	// Add return from main method
	m_pMainMethodCode->code.push_back(_return);

	return true;
}

}}