#pragma once
#include "AG_CodeGenerator.h"

#include <string>
#include <vector>
using namespace std;

#include <hash_map>
using namespace stdext;

// Disable specific warnings 
#pragma warning(disable : 4267)

namespace TinyPascal { namespace Emitter {
//! Java Bytecode Generator Class
class CAG_JavaBytecodeGenerator : public CAG_CodeGenerator
{
protected:
	//! 1 byte unsigned integer
	union u1 { 
		u1() { b[0]=0; };
		u1(unsigned char c) { b[0]=c; };
		unsigned char b[1]; 
	};

	//! 2 byte unsigned integer
	union u2 {
		u2() { v=0; };
		u2(unsigned short s) { v=s; };
		unsigned short v;
		signed short s;
		unsigned char b[2];
	};

	//! 4 byte unsigned integer
	union u4 {
		u4() { v=0; };
		u4(unsigned int i) { v=i; };
		unsigned int v;
		signed int i;
		float f;
		unsigned char b[4];
	};

	//! cp_info structure as specified by JVM specifications
	struct cp_info
	{
		enum type
		{
			CONSTANT_Utf8				= 1,
			CONSTANT_Class				= 7,
			CONSTANT_Fieldref			= 9,
			CONSTANT_Methodref			= 10,
			CONSTANT_InterfaceMethodref = 11,
			CONSTANT_String				= 8,
			CONSTANT_Integer			= 3,
			CONSTANT_Float				= 4,
			CONSTANT_Long				= 5,
			CONSTANT_Double				= 6,
			CONSTANT_NameAndType		= 12
		};

		//! Tag byte denoting the type
		u1 tag;

		//! Constructor
		cp_info() { tag = 0; };

		//! Virtual Destructor
		virtual ~cp_info() {};
	};

	//! CONSTANT_Integer_Info calss as specified by JVM Specification file
	struct CONSTANT_Integer_Info : cp_info
	{
		//! Integer bytes 
		u4 bytes;

		//! Constructor
		CONSTANT_Integer_Info() { tag = CONSTANT_Integer; };
	};

	//! CONSTANT_Float_Info calss as specified by JVM Specification file
	struct CONSTANT_Float_Info : cp_info
	{
		//! Integer bytes 
		u4 bytes;

		//! Constructor
		CONSTANT_Float_Info() { tag = CONSTANT_Float; };
	};

	//! CONSTANT_Utf8_Info class as specified by JVM Specification file
	struct CONSTANT_Utf8_Info : cp_info
	{
		//! Length of the bytes array
		u2 length;
		
		//! Bytes array containing the string
		string bytes;
		
		//! Constructor
		CONSTANT_Utf8_Info() { tag = CONSTANT_Utf8; }
		CONSTANT_Utf8_Info(string s) { tag = CONSTANT_Utf8; bytes = s; }
	};

	//! CONSTANT_Class_Info class as specified by JVM Specification file
	struct CONSTANT_Class_Info : cp_info
	{
		//! Name index in the constant_pool table
		u2 name_index;

		//! Default constructor
		CONSTANT_Class_Info() { tag = CONSTANT_Class; };
		CONSTANT_Class_Info(unsigned short v) { tag = CONSTANT_Class; name_index.v = v; };
		CONSTANT_Class_Info(u2 v) { tag = CONSTANT_Class; name_index = v; };
	};

	//! CONSTANT_Methodref_info class as specified by JVM Specification file
	struct CONSTANT_Methodref_Info : cp_info
	{
		//! Index of CONSTANT_Class_Info in constant_pool table
		u2 class_index;

		//! Index of CONSTANT_NameAndType_Info in constant pool table
		u2 name_and_type_index;

		//! Constructor
		CONSTANT_Methodref_Info() { tag = CONSTANT_Methodref; };
	};

	//! CONSTANT_Methodref_info class as specified by JVM Specification file
	struct CONSTANT_Fieldref_Info : cp_info
	{
		//! Index of CONSTANT_Class_Info or CONSTANT_Interface_Info in constant_pool table
		u2 class_index;

		//! Index of CONSTANT_NameAndType_Info in constant pool table
		u2 name_and_type_index;

		//! Constructor
		CONSTANT_Fieldref_Info() { tag = CONSTANT_Fieldref; };
	};

	//! CONSTANT_NameAndType_info class as specified by JVM Specification file
	struct CONSTANT_NameAndType_Info : cp_info 
	{
		//! Index to CONSTANT_Utf8_Info entry in constant pool table
		u2 name_index;

		//! Index to CONSTANT_Utf8_Info entry in constant pool table
		u2 descriptor_index;

		//! Constructor
		CONSTANT_NameAndType_Info() { tag = CONSTANT_NameAndType; };
	};

	//! Exeption table entry as specified by JVM Specs
	struct exception_table_entry
	{
		//! Starting pc
		u2 start_pc;

		//! Ending pc
		u2 end_pc;

		//! Handler pc
		u2 handler_pc;

		//! Catch type
		u2 catch_type;

		//! Return this structure as bytecode
		string GetBytecode() 
		{
			string strRes; 
			strRes.push_back(start_pc.b[1]);
			strRes.push_back(start_pc.b[0]);
			strRes.push_back(end_pc.b[1]);
			strRes.push_back(end_pc.b[0]);
			strRes.push_back(handler_pc.b[1]);
			strRes.push_back(handler_pc.b[0]);
			strRes.push_back(catch_type.b[1]);
			strRes.push_back(catch_type.b[0]);
			return strRes; 
		};
	};

	//! attribute_info as specified by JVM Specs
	struct attribute_info 
	{
		//! Name index in constant pool table
		u2 attribute_name_index;

		//! attribute length in bytes not including first six bytes of this structure
		u4 attribute_length;

		//! Virtual destructor
		virtual ~attribute_info() {};

		//! Return this structure in bytecode
		virtual string GetBytecode() 
		{
			string strRes;
			strRes.push_back(attribute_name_index.b[1]);
			strRes.push_back(attribute_name_index.b[0]);
			return strRes; 
		};
	};

	//! Code_Attribute as specified in JVM Specs
	struct Code_attribute : attribute_info 
	{
		//! Maximum depth of the operand stack
		u2 max_stack;

		//! Number of local variables
		u2 max_locals;

		//! Instruction code length in bytes
		u4 code_length;

		//! Array containing the instructions of this method
		vector<u1> code;

		//! Number of exeptions table entries
		u2 exception_table_length;

		//! Exception entries
		vector<exception_table_entry> exeption_table;
		
		//! Number of attributes in attribute entry pool
		u2 attributes_count;

		//! Attribute entries
		vector<attribute_info*> attributes;

		//! Clean up in the destructor
		~Code_attribute() 
		{
			for(size_t i=0; i<attributes.size(); ++i)
				delete attributes[i];
		};

		//! Return this structure in bytecode
		string GetBytecode() 
		{
			string strRes;
			strRes.push_back(attribute_name_index.b[1]);
			strRes.push_back(attribute_name_index.b[0]);

			string strTmp;
			strTmp.push_back(max_stack.b[1]);
			strTmp.push_back(max_stack.b[0]);
			strTmp.push_back(max_locals.b[1]);
			strTmp.push_back(max_locals.b[0]);
			code_length.v = code.size();
			strTmp.push_back(code_length.b[3]);
			strTmp.push_back(code_length.b[2]);
			strTmp.push_back(code_length.b[1]);
			strTmp.push_back(code_length.b[0]);
			for(size_t i=0; i<code.size(); ++i)
				strTmp.push_back(code[i].b[0]);
			exception_table_length.v = exeption_table.size();
			strTmp.push_back(exception_table_length.b[1]);
			strTmp.push_back(exception_table_length.b[0]);
			for(size_t i=0; i<exeption_table.size(); ++i)
				strTmp += exeption_table[i].GetBytecode();
			attributes_count.v = attributes.size();
			strTmp.push_back(attributes_count.b[1]);
			strTmp.push_back(attributes_count.b[0]);
			for(size_t i=0; i<attributes.size(); ++i)
				strTmp += attributes[i]->GetBytecode();
			attribute_length.v = strTmp.size();
			strRes.push_back(attribute_length.b[3]);
			strRes.push_back(attribute_length.b[2]);
			strRes.push_back(attribute_length.b[1]);
			strRes.push_back(attribute_length.b[0]);

			strRes += strTmp;
			return strRes; 
		};
	};


	//! method_info as specified by JVM Specs
	struct method_info 
	{
		//! access flags
		u2 access_flags;

		//! name index referring to CONSTANT_Utf8_Info in constant_pool table representing method name
		u2 name_index;

		//! valid method descriptor index in constant_pool table in form CONSTANT_Utf8_Info  
		u2 descriptor_index;

		//! attributes count -> derived from attributes vector
		u2 attributes_count;

		//! vector of attribute_info pointers
		vector<attribute_info*> attributes;

		//! Destructor
		~method_info()
		{
			for(size_t i=0; i<attributes.size(); ++i)
				delete attributes[i];
		};
	};

	//! field_info as specified by JVM Specs
	struct field_info 
	{
		//! Access flags 
		u2 access_flags;

		//! Index in the constant pool table in CONSTANT_Utf8_Info format
		u2 name_index;

		//! Index in the constant_pool table in CONSTANT_Utf8_Info format 
		u2 descriptor_index;

		//! Field attributes count -> derived from attributes vector
		u2 attributes_count;

		//! Vector of pointers to additional attibutes of the filed
		vector<attribute_info*> attributes;

		//! Destructor
		~field_info()
		{
			for(size_t i=0; i<attributes.size(); ++i)
				delete attributes[i];
		};
	};

	//! Instruction set currently supported
	enum Instructions
	{
		_return			= 0xb1,
		getstatic		= 0xb2,
		putstatic		= 0xb3,
		invokevirtual	= 0xb6,
		invokespecial	= 0xb7,
		aload_0			= 0x2a,
		ldc_w			= 0x13,
		f2i				= 0x8b,
		i2f				= 0x86,
		fmul			= 0x6a,
		fdiv			= 0x6e,
		fadd			= 0x62,
		fsub			= 0x66,
		imul			= 0x68,
		idiv			= 0x6c,
		iadd			= 0x60,
		isub			= 0x64,
		if_icmplt		= 0xa1,
		if_icmpgt		= 0xa3,
		iconst_m1		= 0x02,
		iconst_0		= 0x03,
		iconst_1		= 0x04,
		iconst_2		= 0x05,
		_goto			= 0xa7
	};

	//! hash table of UTF8 strings to be reused by constant pool
	hash_map<string, u2> m_Utf8_Table;

	//! hash table of Class_Info structures that need to be reused
	hash_map<string, u2> m_Class_Table;

	//! hash table mapping fields (variables) to index of their references for later use
	hash_map<string, u2> m_Field_Table;

	//! hash table mapping integers factors to indexes in the table
	hash_map<signed int, u2> m_Integer_Table;

	//! hash table mapping integers factors to indexes in the table
	hash_map<float, u2> m_Float_Table;

	//! Index of reference to field referencing static System.out object
	u2 m_PrintStreamRef;

	//! Index of method reference to println-overload that requires no parameters
	u2 m_PrintlnRef;

	//! Index of method reference to println-overload that requires int parameter
	u2 m_PrintlnIntRef;

	//! Index of method reference to println-overload that requires float parameter
	u2 m_PrintlnFloatRef;

	//! Pointer to main method code section, where everything happens
	Code_attribute* m_pMainMethodCode;

	//! Class name
	string m_strClassName;

	//! Indeces adjusted based on endianess of the machine
	unsigned int b0;
	unsigned int b1;
	unsigned int b2;
	unsigned int b3;

	///////////////////////////////////////////////////////////////////////////////////////////////
	// Class File Structure Members

	//! This is java bytecode signature
	static const u4 magic;

	//! Version info
	u2 minor_version;
	u2 major_version;

	//! Constant pool count -> determined by constant pool vector size
	u2 constant_pool_count;

	//! Constant pool is simply vector of cp_info pointers
	vector<cp_info*> constant_pool;

	//! Access flags
	u2 access_flags;

	//! This class index
	u2 this_class;

	//! Super class index
	u2 super_class;

	//! Interfaces count -> determined from interfaces vector
	u2 interfaces_count;

	//! Interfaces index pool
	vector<u2> interfaces;

	//! Fields count -> determined from fields vector
	u2 fields_count;

	//! Vector fields structures for this class
	vector<field_info*> fields;

	//! Method counts -> determined from methods vector
	u2 methods_count;

	//! Methods vector for this class
	vector<method_info*> methods;

	//! Attributes count - derived from attributes vector
	u2 attributes_count;

	//! Vector of attributes
	vector<attribute_info*> attributes;

	///////////////////////////////////////////////////////////////////////////////////////////////
	// Emit structures

	//! Create default class constructor
	void AddDefaultConstructor();

	//! Adds main method 
	void AddMainMethod();

	//! add UTF8 structure and return index in constant_pool
	u2 AddUtf8Info(string bytes);

	//! Add Class info and return index in constant_pool
	u2 AddClassInfo(string name);

	//! Add Name and Type info structure to constant_pool
	u2 AddNameAndTypeInfo(string name, string descriptor);

	//! Add Method reference info to constant_pool
	u2 AddMethodrefInfo(string _class, string name, string type);

	//! Add Field reference info to constant_pool
	u2 AddFieldrefInfo(string _class, string name, string type);

	//! Add integer info reference
	u2 AddIntegerInfo(signed int i);

	//! Add integer info reference
	u2 AddFloatInfo(float f);

	//! Add field to field definition
	void AddField(string name, string type);

	//! Emit u1 to string
	inline void Emit(u1 v, string& program)
	{ program.push_back(v.b[0]); };

	//! Emit u2 to string
	inline void Emit(u2 v, string& program)
	{
		program.push_back(v.b[b0 & 0x01]);
		program.push_back(v.b[b1 & 0x01]);
	};

	//! Emit u4 to string
	inline void Emit(u4 v, string& program)
	{
		program.push_back(v.b[b0]);
		program.push_back(v.b[b1]);
		program.push_back(v.b[b2]);
		program.push_back(v.b[b3]);
	};

	//! Emit u1 to vector
	inline void Emit(u1 v, vector<u1>& program)
	{ program.push_back(v.b[0]); };

	//! Emit u2 to vector
	inline void Emit(u2 v, vector<u1>& program)
	{
		program.push_back(v.b[b0 & 0x01]);
		program.push_back(v.b[b1 & 0x01]);
	};

	//! Emit u4 to vector
	inline void Emit(u4 v, vector<u1>& program)
	{
		program.push_back(v.b[b0]);
		program.push_back(v.b[b1]);
		program.push_back(v.b[b2]);
		program.push_back(v.b[b3]);
	};

	//! Emit cp_info
	void GetBytecode(cp_info* p, string& program);

	//! Emit CONSTANT_Class_Info
	void GetBytecode(CONSTANT_Class_Info* p, string& program);

	//! Emit CONSTANT_Methodref_Info
	void GetBytecode(CONSTANT_Methodref_Info* p, string& program);

	//! Emit CONSTANT_Fieldref_Info
	void GetBytecode(CONSTANT_Fieldref_Info* p, string& program);

	//! Emit CONSTANT_NameAndType_Info
	void GetBytecode(CONSTANT_NameAndType_Info* p, string& program);

	//! Emit CONSTANT_Utf8_Info
	void GetBytecode(CONSTANT_Utf8_Info* p, string& program);

	//! Emit CONSTANT_Integer_Info
	void GetBytecode(CONSTANT_Integer_Info* p, string& program);

	//! Emit CONSTANT_Float_Info
	void GetBytecode(CONSTANT_Float_Info* p, string& program);

	//! Emit method_info structure
	void GetBytecode(method_info* p, string& program);

	//! Emit field_info structure
	void GetBytecode(field_info* p, string& program);

	//! Initializes class file
	void Initialize();

	//! Stack handeling routine -> computes max stack
	inline void ComputeStack(bool bPush, u2& max_stack, bool bReset = false)
	{
		static unsigned int stack_count = 0;
		if(bPush) ++stack_count; else --stack_count;
		if(bReset) stack_count = 0;
		if(max_stack.v<stack_count) max_stack = stack_count;
	}

	///////////////////////////////////////////////////////////////////////////////////////////////
	// Emit parst of AST -> this part is simply converting AST into Java Class File structure

	//! Emits variable declaration part
	bool EmitVarDecl(CAG_TPSymbolTable::SymbolTableType& SymbolTable);

	//! Emits main statement part
	bool EmitMainStatementPart(CAG_TPSymbolTable::TreeChildrenType& MainStatementPart);

	//! Emits assignment operator
	bool EmitAssignment(CAG_TPSymbolTable::TreeChildrenType& OpAssign);

	//! Emits expression 
	bool EmitExpression(CAG_TPSymbolTable::TreeNodeType& Expression);

	//! Emits term
	bool EmitTerm(CAG_TPSymbolTable::TreeNodeType& Term);

	//! Emits addition
	bool EmitAddition(CAG_TPSymbolTable::TreeNodeType& Addition);

	//! Emits addition
	bool EmitSubtraction(CAG_TPSymbolTable::TreeNodeType& Subtraction);

	//! Emits division
	bool EmitDivision(CAG_TPSymbolTable::TreeNodeType& Division);

	//! Emits multiplication
	bool EmitMultiplication(CAG_TPSymbolTable::TreeNodeType& Multiplication);

	//! Emits Identifier by pushing it onto stack
	bool PushIdentifier(CAG_TPSymbolTable::TreeNodeType& Identifier);

	//! Emits Integer Factor by pushing it onto stack
	bool PushIntegerFactor(CAG_TPSymbolTable::TreeNodeType& Factor);

	//! Emits Real Factor by pushing it onto stack
	bool PushRealFactor(CAG_TPSymbolTable::TreeNodeType& Factor);

	//! Emits Writeln Statement
	bool EmitWritelnStatement(CAG_TPSymbolTable::TreeChildrenType& WritelnStmt);

	//! Emits FOR loop
	bool EmitForLoop(CAG_TPSymbolTable::TreeChildrenType& ForLoop);

public:
	//! Default constructor
	CAG_JavaBytecodeGenerator(void);

	//! Destructor
	~CAG_JavaBytecodeGenerator(void);

	//! Emit Java bytecode code
	bool Emit(CAG_TPSymbolTable::TreeChildrenType& AST,
			  CAG_TPSymbolTable::SymbolTableType& SymbolTable);

	//! Returns this class as bytecode
	string GetProgram();

	//! Returns name of the file that needs to be created
	string GetFileName() { return m_strClassName + ".class"; };
};
}}
