/*	\file Grammar.h
	\author Amer Gerzic
	\description

	Grammar.h contains grammar definition for TinyPascal language defined
	using Spirit 1.8.3 parser framework.
*/

#pragma once

#include "stdafx.h"

#include <boost/spirit/core.hpp>
#include <boost/spirit/symbols/symbols.hpp>
#include <boost/spirit/tree/ast.hpp>
#include <boost/spirit/tree/parse_tree.hpp>
#include <boost/spirit/iterator/position_iterator.hpp>
#include <boost/spirit/utility/distinct.hpp>
using namespace boost::spirit;

namespace TinyPascal {
	namespace Parser {

const distinct_directive<> keyword_d("0-9a-zA-Z_");
const distinct_directive<> number_d("a-zA-Z_");

//! Grmmar used for parsing language
class CAG_TinyPascalGrammar : public grammar<CAG_TinyPascalGrammar>
{
public:
	//! Define constants for each node
	static const int ProgramID					= 1;
	static const int ProgramHeadingID			= 2;
	static const int GlobalDeclarationPartID	= 3;
	static const int MainStatementPartID		= 4;
	static const int IdentifierID				= 5;
	static const int VarDeclarationID			= 6;
	static const int VarDeclID					= 7;
	static const int RealFactorID				= 8;
	static const int TypeID						= 9;
	static const int ForStatementID				= 10;
	static const int StatementID				= 11;
	static const int ExpressionID				= 12;
	static const int WritelnStmtID				= 13;
	static const int OpAssignID					= 14;
	static const int TermID						= 15;
	static const int FactorID					= 16;
	static const int IntegerFactorID			= 17;

	//! Return rule name from rule ID
	static string GetRuleName(const int RuleID)
	{
		string Res = "{KEYWORD}";
		switch(RuleID)
		{
		case ProgramID:
			Res = "{Program}";
			break;
		case ProgramHeadingID:
			Res = "{ProgramHeading}";
			break;
		case GlobalDeclarationPartID:
			Res = "{GlobalDeclarationPart}";
			break;
		case MainStatementPartID:
			Res = "{MainStatementPart}";
			break;
		case IdentifierID:
			Res = "{Identifier}";
			break;
		case VarDeclarationID:
			Res = "{VarDeclaration}";
			break;
		case VarDeclID:
			Res = "{VarDecl}";
			break;
		case RealFactorID:
			Res = "{Real}";
			break;
		case TypeID:
			Res = "{Type}";
			break;
		case ForStatementID:
			Res = "{ForStatement}";
			break;
		case StatementID:
			Res = "{Statement}";
			break;
		case ExpressionID:
			Res = "{Expression}";
			break;
		case WritelnStmtID:
			Res = "{WritelnStatement}";
			break;
		case OpAssignID:
			Res = "{OpAssign}";
			break;
		case TermID:
			Res = "{Term}";
			break;
		case FactorID:
			Res = "{Factor}";
			break;
		case IntegerFactorID:
			Res = "{Integer}";
			break;
		}

		return Res;
	}

	//! Default constructor
	CAG_TinyPascalGrammar() {}

	//! Define nested class as required by Spirit framework
	template <typename ScannerT> class definition
	{
	public:
		//! Parameterized constructor defines grammar
		definition(CAG_TinyPascalGrammar const& /*self*/) 
		{
			#ifdef BOOST_SPIRIT_DEBUG
            BOOST_SPIRIT_DEBUG_RULE(program);
			BOOST_SPIRIT_DEBUG_RULE(program_heading);
			BOOST_SPIRIT_DEBUG_RULE(global_declaration_part);
			BOOST_SPIRIT_DEBUG_RULE(main_statement_part);
			BOOST_SPIRIT_DEBUG_RULE(identifier);
			BOOST_SPIRIT_DEBUG_RULE(var_declaration);
			BOOST_SPIRIT_DEBUG_RULE(var_decl);
			BOOST_SPIRIT_DEBUG_RULE(type);
			BOOST_SPIRIT_DEBUG_RULE(statement);
			BOOST_SPIRIT_DEBUG_RULE(expression);
			BOOST_SPIRIT_DEBUG_RULE(writeln_stmt);
			BOOST_SPIRIT_DEBUG_RULE(term);
			BOOST_SPIRIT_DEBUG_RULE(factor);
			BOOST_SPIRIT_DEBUG_RULE(op_assign);
			BOOST_SPIRIT_DEBUG_RULE(for_statement);
            #endif

			/**************************************************************
			* Keyword Helpers (This is used to avoid confusion between 
			* a identifier and a keyword.)
			***************************************************************/
			keywords
				=	_T("program"), _T("begin"), _T("end"), _T("var"),
					_T("integer"), _T("real"), _T("writeln"), _T("for"),
					_T("to"), _T("downto"), _T("do");

			/**************************************************************
			* Token Helpers (Single Characters)
			***************************************************************/
			chlit<TCHAR> DOT('.');
			chlit<TCHAR> SEMI(';');
			chlit<TCHAR> COLON(':');
			chlit<TCHAR> PLUS('+');
			chlit<TCHAR> MINUS('-');
			chlit<TCHAR> MULT('*');
			chlit<TCHAR> DIV('/');
			chlit<TCHAR> EQUALS('=');
			chlit<TCHAR> LPAREN('(');
			chlit<TCHAR> RPAREN(')');
			chlit<TCHAR> COMMA(',');

			/**************************************************************
			* Token Helpers (Multiple Characters)
			***************************************************************/
			inhibit_case<strlit<> > ASSIGN	= as_lower_d[":="];
			inhibit_case<strlit<> > PROGRAM	= as_lower_d["program"];
			inhibit_case<strlit<> > BEGIN	= as_lower_d["begin"];
			inhibit_case<strlit<> > END		= as_lower_d["end"];
			inhibit_case<strlit<> > VAR		= as_lower_d["var"];
			inhibit_case<strlit<> > INTEGER	= as_lower_d["integer"];
			inhibit_case<strlit<> > REAL	= as_lower_d["real"];
			inhibit_case<strlit<> > WRITELN	= as_lower_d["writeln"];
			inhibit_case<strlit<> > FOR		= as_lower_d["for"];
			inhibit_case<strlit<> > TO		= as_lower_d["to"];
			inhibit_case<strlit<> > DOWNTO	= as_lower_d["downto"];
			inhibit_case<strlit<> > DO		= as_lower_d["do"];

			/**************************************************************
			* Productions
			***************************************************************/
			program
				=	program_heading >>
					!global_declaration_part >>
					main_statement_part >> 
					no_node_d[DOT];

			program_heading
				=	no_node_d[keyword_d[PROGRAM]] >> identifier >> no_node_d[SEMI];

			global_declaration_part
				=	var_declaration;

			var_declaration
				=	keyword_d[VAR] >> +var_decl;

			var_decl
				=	identifier  >> 
					*(no_node_d[COMMA] >> identifier) >> 
					no_node_d[COLON] >> 
					root_node_d[type] >> 
					no_node_d[SEMI];

			type
				=	INTEGER | REAL;
			
			main_statement_part
				=	keyword_d[BEGIN] >> 
						statement >> *(no_node_d[SEMI] >> statement) >> 
					keyword_d[END];

			op_assign
				=	identifier >> root_node_d[ASSIGN] >> expression;

			statement
				=	main_statement_part |
					op_assign |
					writeln_stmt |
					for_statement |
					no_node_d[epsilon_p];

			for_statement
				=	keyword_d[FOR] >> op_assign >> (keyword_d[TO] | keyword_d[DOWNTO]) >> expression >> keyword_d[DO] >> statement;

			writeln_stmt
				=	root_node_d[WRITELN] >> 
					((no_node_d[LPAREN] >> 
						expression >> *(no_node_d[COMMA] >> expression) >> 
					no_node_d[RPAREN]) | 
					no_node_d[epsilon_p]);

			expression
				=	term >>
					*((root_node_d[PLUS] >> term) | 
					  (root_node_d[MINUS] >> term));

			term 
				=	factor >>
					*((root_node_d[MULT] >> factor) |
					  (root_node_d[DIV] >> factor));

			real_factor
				=	number_d[strict_real_p];

			integer_factor
				=	number_d[int_p];

			factor
				=	identifier | 
					real_factor |
					integer_factor |
					inner_node_d[LPAREN >> expression >> RPAREN];

			identifier
				=	token_node_d[
						lexeme_d[
							(alpha_p >> *(alnum_p)) - (keywords >> anychar_p - alnum_p)
					]];
		}

		//! Keyword helper
		symbols<int, TCHAR> keywords;

		//! Productions
		rule<ScannerT, parser_context<>, parser_tag<ProgramID> > program;
		rule<ScannerT, parser_context<>, parser_tag<ProgramHeadingID> > program_heading;
		rule<ScannerT, parser_context<>, parser_tag<GlobalDeclarationPartID> > global_declaration_part;
		rule<ScannerT, parser_context<>, parser_tag<MainStatementPartID> > main_statement_part;
		rule<ScannerT, parser_context<>, parser_tag<IdentifierID> > identifier;
		rule<ScannerT, parser_context<>, parser_tag<VarDeclarationID> > var_declaration;
		rule<ScannerT, parser_context<>, parser_tag<VarDeclID> > var_decl;
		rule<ScannerT, parser_context<>, parser_tag<TypeID> > type;
		rule<ScannerT, parser_context<>, parser_tag<StatementID> > statement;
		rule<ScannerT, parser_context<>, parser_tag<ExpressionID> > expression;
		rule<ScannerT, parser_context<>, parser_tag<WritelnStmtID> > writeln_stmt; 
		rule<ScannerT, parser_context<>, parser_tag<TermID> > term;
		rule<ScannerT, parser_context<>, parser_tag<FactorID> > factor;
		rule<ScannerT, parser_context<>, parser_tag<OpAssignID> > op_assign;
		rule<ScannerT, parser_context<>, parser_tag<ForStatementID> > for_statement;
		rule<ScannerT, parser_context<>, parser_tag<RealFactorID> > real_factor;
		rule<ScannerT, parser_context<>, parser_tag<IntegerFactorID> > integer_factor;

		//! Override start function to return starting production
		rule<ScannerT, parser_context<>, parser_tag<ProgramID> > const& start() const { return program; }
	};
};

/*! 
	Grammar used for parsing comments (instruction to scanner on what to skip).
	Skip grammar is equivalent to ignored token in SableCC.
*/
class CAG_TinyPascalSkipper : public grammar<CAG_TinyPascalSkipper>
{
public:
	//! Default constructor
	CAG_TinyPascalSkipper() {}

	//! Define nested class as required by Spirit framework
	template <typename ScannerT> class definition
	{
	public:
		//! Parameterized constructor defines grammar
		definition(CAG_TinyPascalSkipper const& /*self*/) 
		{
			/*
				Here we want to instruct parser to skip any type of space by using
				space_p rule that is built into framework and maches space, tab, return, 
				and newline. Also using alternative we are specifying that we want 
				to skip any comment within curly brackets {}. anychar_p is built in 
				rule that parses any character respectivly. It is equivalent to DOT 
				operator for Regular Expressions.
			*/
			Skip 
				=	space_p | '{' >> (*(anychar_p - '}')) >> '}';

			#ifdef BOOST_SPIRIT_DEBUG
            BOOST_SPIRIT_DEBUG_RULE(Skip);
			#endif
		}

		//! Starting production
		rule<ScannerT> Skip;

		//! Override start function to return starting production
		rule<ScannerT> const& start() const { return Skip; }
	};
};
}}